@@ -1,5 +1,5 @@
 /* Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
-   Copyright (c) 2016, MariaDB
+   Copyright (c) 2016, 2020, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -16,7 +16,6 @@
 
 
 #include "sql_plugin.h"
-#include "hash.h"
 #include "table.h"
 #include "rpl_gtid.h"
 #include "sql_class.h"
@@ -24,7 +23,7 @@
 #include "sql_plugin.h"
 #include "set_var.h"
 
-void State_tracker::mark_as_changed(THD *thd, LEX_CSTRING *tracked_item_name)
+void State_tracker::set_changed(THD *thd)
 {
   m_changed= true;
   thd->lex->safe_to_cache_query= 0;
@@ -73,7 +72,8 @@ void Session_sysvars_tracker::vars_list::copy(vars_list* from, THD *thd)
 bool Session_sysvars_tracker::vars_list::insert(const sys_var *svar)
 {
   sysvar_node_st *node;
-  if (!(node= (sysvar_node_st *) my_malloc(sizeof(sysvar_node_st),
+  if (!(node= (sysvar_node_st *) my_malloc(PSI_INSTRUMENT_ME,
+                                           sizeof(sysvar_node_st),
                                            MYF(MY_WME |
                                                (mysqld_server_initialized ?
                                                 MY_THREAD_SPECIFIC : 0)))))
@@ -327,7 +327,8 @@ void Session_sysvars_tracker::init(THD *thd)
               global_system_variables.session_track_system_variables);
   DBUG_ASSERT(global_system_variables.session_track_system_variables);
   thd->variables.session_track_system_variables=
-    my_strdup(global_system_variables.session_track_system_variables,
+    my_strdup(PSI_INSTRUMENT_ME,
+              global_system_variables.session_track_system_variables,
               MYF(MY_WME | MY_THREAD_SPECIFIC));
 }
 
@@ -384,11 +385,11 @@ bool Session_sysvars_tracker::update(THD *thd, set_var *var)
   size_t length= 1;
 
   if (var->save_result.string_value.str)
-    copy= my_memdup(var->save_result.string_value.str,
+    copy= my_memdup(PSI_INSTRUMENT_ME, var->save_result.string_value.str,
                     (length= var->save_result.string_value.length + 1),
                     MYF(MY_WME | MY_THREAD_SPECIFIC));
     else
-      copy= my_strdup("", MYF(MY_WME | MY_THREAD_SPECIFIC));
+      copy= my_strdup(PSI_INSTRUMENT_ME, "", MYF(MY_WME | MY_THREAD_SPECIFIC));
 
   if (!copy)
     return true;
@@ -506,11 +507,12 @@ bool Session_sysvars_tracker::store(THD *thd, String *buf)
   @param               [IN] pointer on a variable
 */
 
-void Session_sysvars_tracker::mark_as_changed(THD *thd,
-                                              LEX_CSTRING *var)
+void Session_sysvars_tracker::mark_as_changed(THD *thd, const sys_var *var)
 {
   sysvar_node_st *node;
-  sys_var *svar= (sys_var *)var;
+
+  if (!is_enabled())
+    return;
 
   if (!m_parsed)
   {
@@ -529,10 +531,10 @@ void Session_sysvars_tracker::mark_as_changed(THD *thd,
     Check if the specified system variable is being tracked, if so
     mark it as changed and also set the class's m_changed flag.
   */
-  if (orig_list.is_enabled() && (node= orig_list.insert_or_search(svar)))
+  if (orig_list.is_enabled() && (node= orig_list.insert_or_search(var)))
   {
     node->m_changed= true;
-    State_tracker::mark_as_changed(thd, var);
+    set_changed(thd);
   }
 }
 
@@ -679,7 +681,7 @@ bool Transaction_state_tracker::update(THD *thd, set_var *)
     }
     if (thd->variables.session_track_transaction_info == TX_TRACK_CHISTICS)
       tx_changed       |= TX_CHG_CHISTICS;
-    mark_as_changed(thd, NULL);
+    set_changed(thd);
   }
   else
     m_enabled= false;
@@ -1112,7 +1114,7 @@ void Transaction_state_tracker::set_read_flags(THD *thd,
   {
     tx_read_flags = flags;
     tx_changed   |= TX_CHG_CHISTICS;
-    mark_as_changed(thd, NULL);
+    set_changed(thd);
   }
 }
 
@@ -1131,7 +1133,7 @@ void Transaction_state_tracker::set_isol_level(THD *thd,
   {
     tx_isol_level = level;
     tx_changed   |= TX_CHG_CHISTICS;
-    mark_as_changed(thd, NULL);
+    set_changed(thd);
   }
 }
 
@@ -1181,6 +1183,43 @@ bool Session_state_change_tracker::store(THD *thd, String *buf)
   return false;
 }
 
+
+bool User_variables_tracker::update(THD *thd, set_var *)
+{
+  m_enabled= thd->variables.session_track_user_variables;
+  return false;
+}
+
+
+bool User_variables_tracker::store(THD *thd, String *buf)
+{
+  for (ulong i= 0; i < m_changed_user_variables.size(); i++)
+  {
+    auto var= m_changed_user_variables.at(i);
+    String value_str;
+    bool null_value;
+
+    var->val_str(&null_value, &value_str, DECIMAL_MAX_SCALE);
+
+    size_t length= net_length_size(var->name.length) + var->name.length;
+    if (!null_value)
+      length+= net_length_size(value_str.length()) + value_str.length();
+
+    if (buf->reserve(sizeof(char) + length + net_length_size(length)))
+      return true;
+
+    buf->q_append(static_cast<char>(SESSION_TRACK_USER_VARIABLES));
+    buf->q_net_store_length(length);
+    buf->q_net_store_data(reinterpret_cast<const uchar*>(var->name.str),
+                          var->name.length);
+    if (!null_value)
+      buf->q_net_store_data(reinterpret_cast<const uchar*>(value_str.ptr()),
+                            value_str.length());
+  }
+  m_changed_user_variables.clear();
+  return false;
+}
+
 ///////////////////////////////////////////////////////////////////////////////
 
 /**
@@ -1225,18 +1264,27 @@ void Session_tracker::store(THD *thd, String *buf)
   }
 
   size_t length= buf->length() - start;
-  uchar *data= (uchar *)(buf->ptr() + start);
+  uchar *data;
   uint size;
 
   if ((size= net_length_size(length)) != 1)
   {
-    if (buf->reserve(size - 1, EXTRA_ALLOC))
+    if (buf->reserve(size - 1, 0))
     {
       buf->length(start); // it is safer to have 0-length block in case of error
       return;
     }
+
+    /*
+      The 'buf->reserve()' can change the buf->ptr() so we cannot
+      calculate the 'data' earlier.
+    */
+    buf->length(buf->length() + (size - 1));
+    data= (uchar *)(buf->ptr() + start);
     memmove(data + (size - 1), data, length);
   }
+  else
+    data= (uchar *)(buf->ptr() + start);
 
   net_store_length(data - 1, length);
 }
