@@ -23,6 +23,7 @@
 #include <sql_audit.h>
 #include <debug_sync.h>
 #include <threadpool.h>
+
 #ifdef WITH_WSREP
 #include "wsrep_trans_observer.h"
 #endif /* WITH_WSREP */
@@ -38,6 +39,8 @@ uint threadpool_max_threads;
 uint threadpool_oversubscribe;
 uint threadpool_mode;
 uint threadpool_prio_kickup_timer;
+my_bool threadpool_exact_stats;
+my_bool threadpool_dedicated_listener;
 
 /* Stats */
 TP_STATISTICS tp_stats;
@@ -93,7 +96,7 @@ struct Worker_thread_context
   {
     PSI_CALL_set_thread(psi_thread);
     set_mysys_var(mysys_var);
-    pthread_setspecific(THR_THD, 0);
+    set_current_thd(nullptr);
   }
 };
 
@@ -146,7 +149,7 @@ static void thread_attach(THD* thd)
   set_mysys_var(thd->mysys_var);
   thd->thread_stack=(char*)&thd;
   thd->store_globals();
-  PSI_CALL_set_thread(thd->event_scheduler.m_psi);
+  PSI_CALL_set_thread(thd->get_psi());
   mysql_socket_set_thread_owner(thd->net.vio->mysql_socket);
 }
 
@@ -159,9 +162,8 @@ static TP_PRIORITY get_priority(TP_connection *c)
   DBUG_ASSERT(c->thd == current_thd);
   TP_PRIORITY prio= (TP_PRIORITY)c->thd->variables.threadpool_priority;
   if (prio == TP_PRIORITY_AUTO)
-  {
-    return c->thd->transaction.is_active() ? TP_PRIORITY_HIGH : TP_PRIORITY_LOW;
-  }
+    prio= c->thd->transaction.is_active() ? TP_PRIORITY_HIGH : TP_PRIORITY_LOW;
+
   return prio;
 }
 
@@ -231,21 +233,13 @@ static THD* threadpool_add_connection(CONNECT *connect, void *scheduler_data)
   set_mysys_var(NULL);
   my_thread_init();
   st_my_thread_var* mysys_var= my_thread_var;
+  PSI_CALL_set_thread(PSI_CALL_new_thread(key_thread_one_connection, connect, 0));
   if (!mysys_var ||!(thd= connect->create_thd(NULL)))
   {
     /* Out of memory? */
     connect->close_and_delete();
     if (mysys_var)
-    {
-#ifdef HAVE_PSI_INTERFACE
-      /*
-       current PSI is still from worker thread.
-       Set to 0, to avoid premature cleanup by my_thread_end
-      */
-      if (PSI_server) PSI_server->set_thread(0);
-#endif
       my_thread_end();
-    }
     return NULL;
   }
   delete connect;
@@ -253,11 +247,6 @@ static THD* threadpool_add_connection(CONNECT *connect, void *scheduler_data)
   thd->set_mysys_var(mysys_var);
   thd->event_scheduler.data= scheduler_data;
 
-  /* Create new PSI thread for use with the THD. */
-  thd->event_scheduler.m_psi=
-    PSI_CALL_new_thread(key_thread_one_connection, thd, thd->thread_id);
-
-
   /* Login. */
   thread_attach(thd);
   re_init_net_server_extension(thd);
@@ -266,8 +255,7 @@ static THD* threadpool_add_connection(CONNECT *connect, void *scheduler_data)
   thd->start_utime= now;
   thd->thr_create_utime= now;
 
-  if (setup_connection_thread_globals(thd))
-    goto end;
+  setup_connection_thread_globals(thd);
 
   if (thd_prepare_connection(thd))
     goto end;
@@ -297,6 +285,7 @@ static void threadpool_remove_connection(THD *thd)
   end_connection(thd);
   close_connection(thd, 0);
   unlink_thd(thd);
+  PSI_CALL_delete_current_thread(); // before THD is destroyed
   delete thd;
 
   /*
@@ -387,19 +376,6 @@ end:
 }
 
 
-
-/* Dummy functions, do nothing */
-
-static bool tp_init_new_connection_thread()
-{
-  return 0;
-}
-
-static bool tp_end_thread(THD *, bool)
-{
-  return 0;
-}
-
 static TP_pool *pool;
 
 static bool tp_init()
@@ -475,34 +451,25 @@ void tp_timeout_handler(TP_connection *c)
 {
   if (c->state != TP_STATE_IDLE)
     return;
-  THD *thd= c->thd;
+  THD *thd=c->thd;
   mysql_mutex_lock(&thd->LOCK_thd_kill);
-  Vio *vio= thd->net.vio;
-  if (vio && (vio_pending(vio) > 0 || vio->has_data(vio)) &&
-      c->state == TP_STATE_IDLE)
-  {
-    /*
-     There is some data on that connection, i.e
-     i.e there was no inactivity timeout.
-     Don't kill.
-    */
-    c->state= TP_STATE_PENDING;
-  }
-  else if (c->state == TP_STATE_IDLE)
-  {
-    thd->set_killed_no_mutex(KILL_WAIT_TIMEOUT);
-    c->priority= TP_PRIORITY_HIGH;
-    post_kill_notification(thd);
-  }
+  thd->set_killed_no_mutex(KILL_WAIT_TIMEOUT);
+  c->priority= TP_PRIORITY_HIGH;
+  post_kill_notification(thd);
   mysql_mutex_unlock(&thd->LOCK_thd_kill);
 }
 
+MY_ALIGNED(CPU_LEVEL1_DCACHE_LINESIZE) Atomic_counter<unsigned long long> tp_waits[THD_WAIT_LAST];
 
 static void tp_wait_begin(THD *thd, int type)
 {
   TP_connection *c = get_TP_connection(thd);
   if (c)
+  {
+    DBUG_ASSERT(type > 0 && type < THD_WAIT_LAST);
+    tp_waits[type]++;
     c->wait_begin(type);
+  }
 }
 
 
@@ -533,18 +500,16 @@ static scheduler_functions tp_scheduler_functions=
   NULL,
   NULL,
   tp_init,                            // init
-  tp_init_new_connection_thread,      // init_new_connection_thread
   tp_add_connection,                  // add_connection
   tp_wait_begin,                      // thd_wait_begin
   tp_wait_end,                        // thd_wait_end
   tp_post_kill_notification,          // post kill notification
-  tp_end_thread,                      // Dummy function
   tp_end                              // end
 };
 
 void pool_of_threads_scheduler(struct scheduler_functions *func,
     ulong *arg_max_connections,
-    uint *arg_connection_count)
+    Atomic_counter<uint> *arg_connection_count)
 {
   *func = tp_scheduler_functions;
   func->max_threads= threadpool_max_threads;
