@@ -191,32 +191,35 @@ static
 void
 btr_search_check_free_space_in_heap(const dict_index_t* index)
 {
+	hash_table_t*	table;
+	mem_heap_t*	heap;
+
+	ut_ad(!btr_search_own_any(RW_LOCK_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_X));
+
+	table = btr_get_search_table(index);
+
+	heap = table->heap;
+
 	/* Note that we peek the value of heap->free_block without reserving
 	the latch: this is ok, because we will not guarantee that there will
 	be enough free space in the hash table. */
 
-	buf_block_t*	block = buf_block_alloc(NULL);
-	rw_lock_t*	latch = btr_get_search_latch(index);
-	hash_table_t*	table;
-	mem_heap_t*	heap;
-
-	rw_lock_x_lock(latch);
-
-	if (!btr_search_enabled) {
-		goto func_exit;
-	}
-
-	table = btr_get_search_table(index);
-	heap = table->heap;
-
 	if (heap->free_block == NULL) {
-		heap->free_block = block;
-	} else {
-func_exit:
-		buf_block_free(block);
-	}
+		buf_block_t*	block = buf_block_alloc();
+		rw_lock_t*	ahi_latch = btr_get_search_latch(index);
 
-	rw_lock_x_unlock(latch);
+		rw_lock_x_lock(ahi_latch);
+
+		if (btr_search_enabled
+		    && heap->free_block == NULL) {
+			heap->free_block = block;
+		} else {
+			buf_block_free(block);
+		}
+
+		rw_lock_x_unlock(ahi_latch);
+	}
 }
 
 /** Creates and initializes the adaptive search system at a database start.
@@ -244,144 +247,128 @@ void btr_search_sys_create(ulint hash_size)
 	btr_search_sys = reinterpret_cast<btr_search_sys_t*>(
 		ut_malloc(sizeof(btr_search_sys_t), mem_key_ahi));
 
-	btr_search_sys->hash_tables = NULL;
+	btr_search_sys->hash_tables = reinterpret_cast<hash_table_t**>(
+		ut_malloc(sizeof(hash_table_t*) * btr_ahi_parts, mem_key_ahi));
+
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
+
+		btr_search_sys->hash_tables[i] =
+			ib_create((hash_size / btr_ahi_parts),
+				  LATCH_ID_HASH_TABLE_MUTEX,
+				  0, MEM_HEAP_FOR_BTR_SEARCH);
+
+#if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
+		btr_search_sys->hash_tables[i]->adaptive = TRUE;
+#endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
+	}
+}
+
+/** Resize hash index hash table.
+@param[in]	hash_size	hash index hash table size */
+void btr_search_sys_resize(ulint hash_size)
+{
+	/* Step-1: Lock all search latches in exclusive mode. */
+	btr_search_x_lock_all();
 
 	if (btr_search_enabled) {
-		btr_search_enable();
+
+		btr_search_x_unlock_all();
+
+		ib::error() << "btr_search_sys_resize failed because"
+			" hash index hash table is not empty.";
+		ut_ad(0);
+		return;
 	}
+
+	/* Step-2: Recreate hash tables with new size. */
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
+
+		mem_heap_free(btr_search_sys->hash_tables[i]->heap);
+		hash_table_free(btr_search_sys->hash_tables[i]);
+
+		btr_search_sys->hash_tables[i] =
+			ib_create((hash_size / btr_ahi_parts),
+				  LATCH_ID_HASH_TABLE_MUTEX,
+				  0, MEM_HEAP_FOR_BTR_SEARCH);
+
+#if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
+		btr_search_sys->hash_tables[i]->adaptive = TRUE;
+#endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
+	}
+
+	/* Step-3: Unlock all search latches from exclusive mode. */
+	btr_search_x_unlock_all();
 }
 
 /** Frees the adaptive search system at a database shutdown. */
 void btr_search_sys_free()
 {
-  if (!btr_search_sys)
-  {
-    ut_ad(!btr_search_latches);
-    return;
-  }
+	if (!btr_search_sys) {
+		ut_ad(!btr_search_latches);
+		return;
+	}
 
-  ut_ad(btr_search_sys);
-  ut_ad(btr_search_latches);
+	ut_ad(btr_search_latches);
 
-  if (btr_search_sys->hash_tables)
-  {
-    for (ulint i= 0; i < btr_ahi_parts; ++i)
-    {
-      mem_heap_free(btr_search_sys->hash_tables[i]->heap);
-      hash_table_free(btr_search_sys->hash_tables[i]);
-    }
-    ut_free(btr_search_sys->hash_tables);
-  }
+	/* Step-1: Release the hash tables. */
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
 
-  ut_free(btr_search_sys);
-  btr_search_sys= NULL;
+		mem_heap_free(btr_search_sys->hash_tables[i]->heap);
+		hash_table_free(btr_search_sys->hash_tables[i]);
 
-  /* Free all latches. */
-  for (ulint i= 0; i < btr_ahi_parts; ++i)
-  {
-    rw_lock_free(btr_search_latches[i]);
-    ut_free(btr_search_latches[i]);
-  }
+	}
 
-  ut_free(btr_search_latches);
-  btr_search_latches= NULL;
+	ut_free(btr_search_sys->hash_tables);
+	ut_free(btr_search_sys);
+	btr_search_sys = NULL;
+
+	/* Step-2: Release all allocates latches. */
+	for (ulint i = 0; i < btr_ahi_parts; ++i) {
+
+		rw_lock_free(btr_search_latches[i]);
+		ut_free(btr_search_latches[i]);
+	}
+
+	ut_free(btr_search_latches);
+	btr_search_latches = NULL;
 }
 
 /** Set index->ref_count = 0 on all indexes of a table.
 @param[in,out]	table	table handler */
-static void btr_search_disable_ref_count(dict_table_t *table)
+static
+void
+btr_search_disable_ref_count(
+	dict_table_t*	table)
 {
-  for (dict_index_t *index= dict_table_get_first_index(table); index;
-       index= dict_table_get_next_index(index))
-    index->search_info->ref_count= 0;
+	dict_index_t*	index;
+
+	ut_ad(mutex_own(&dict_sys.mutex));
+
+	for (index = dict_table_get_first_index(table);
+	     index != NULL;
+	     index = dict_table_get_next_index(index)) {
+		index->search_info->ref_count = 0;
+	}
 }
 
-/** Lazily free detached metadata when removing the last reference. */
-ATTRIBUTE_COLD static void btr_search_lazy_free(dict_index_t *index)
-{
-  ut_ad(index->freed());
-  dict_table_t *table= index->table;
-  /* Perform the skipped steps of dict_index_remove_from_cache_low(). */
-  UT_LIST_REMOVE(table->freed_indexes, index);
-  rw_lock_free(&index->lock);
-  dict_mem_index_free(index);
-
-  if (!UT_LIST_GET_LEN(table->freed_indexes) &&
-      !UT_LIST_GET_LEN(table->indexes))
-  {
-    ut_ad(table->id == 0);
-    dict_mem_table_free(table);
-  }
-}
-
-/** Clear the adaptive hash index on all pages in the buffer pool. */
-static void buf_pool_clear_hash_index()
-{
-  ut_ad(btr_search_own_all(RW_LOCK_X));
-  ut_ad(!btr_search_enabled);
-
-  std::set<dict_index_t*> garbage;
-
-  for (ulong p = 0; p < srv_buf_pool_instances; p++)
-  {
-    buf_pool_t *buf_pool= buf_pool_from_array(p);
-    buf_chunk_t *chunks= buf_pool->chunks;
-    buf_chunk_t *chunk= chunks + buf_pool->n_chunks;
-
-    while (--chunk >= chunks)
-    {
-      buf_block_t *block= chunk->blocks;
-      for (ulint i= chunk->size; i--; block++)
-      {
-        dict_index_t *index= block->index;
-        assert_block_ahi_valid(block);
-
-        /* We can clear block->index and block->n_pointers when
-        btr_search_own_all(RW_LOCK_X); see the comments in buf0buf.h */
-
-        if (!index)
-        {
-# if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-          ut_a(!block->n_pointers);
-# endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
-          continue;
-        }
-
-        ut_d(buf_page_state state= buf_block_get_state(block));
-        /* Another thread may have set the state to
-        BUF_BLOCK_REMOVE_HASH in buf_LRU_block_remove_hashed().
-
-        The state change in buf_page_realloc() is not observable here,
-        because in that case we would have !block->index.
-
-        In the end, the entire adaptive hash index will be removed. */
-        ut_ad(state == BUF_BLOCK_FILE_PAGE || state == BUF_BLOCK_REMOVE_HASH);
-# if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-        block->n_pointers= 0;
-# endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
-        if (index->freed())
-          garbage.insert(index);
-        block->index= NULL;
-      }
-    }
-  }
-
-  for (std::set<dict_index_t*>::iterator i= garbage.begin();
-       i != garbage.end(); i++)
-    btr_search_lazy_free(*i);
-}
-
-/** Disable the adaptive hash search system and empty the index. */
-void btr_search_disable()
+/** Disable the adaptive hash search system and empty the index.
+@param[in]	need_mutex	need to acquire dict_sys.mutex */
+void btr_search_disable(bool need_mutex)
 {
 	dict_table_t*	table;
 
-	mutex_enter(&dict_sys.mutex);
+	if (need_mutex) {
+		mutex_enter(&dict_sys.mutex);
+	}
 
+	ut_ad(mutex_own(&dict_sys.mutex));
 	btr_search_x_lock_all();
 
 	if (!btr_search_enabled) {
-		mutex_exit(&dict_sys.mutex);
+		if (need_mutex) {
+			mutex_exit(&dict_sys.mutex);
+		}
+
 		btr_search_x_unlock_all();
 		return;
 	}
@@ -402,60 +389,62 @@ void btr_search_disable()
 		btr_search_disable_ref_count(table);
 	}
 
-	mutex_exit(&dict_sys.mutex);
+	if (need_mutex) {
+		mutex_exit(&dict_sys.mutex);
+	}
 
 	/* Set all block->index = NULL. */
-	buf_pool_clear_hash_index();
+	buf_pool.clear_hash_index();
 
 	/* Clear the adaptive hash index. */
 	for (ulint i = 0; i < btr_ahi_parts; ++i) {
-		mem_heap_free(btr_search_sys->hash_tables[i]->heap);
-		hash_table_free(btr_search_sys->hash_tables[i]);
+		hash_table_clear(btr_search_sys->hash_tables[i]);
+		mem_heap_empty(btr_search_sys->hash_tables[i]->heap);
 	}
-	ut_free(btr_search_sys->hash_tables);
-	btr_search_sys->hash_tables = NULL;
 
 	btr_search_x_unlock_all();
 }
 
-/** Enable the adaptive hash search system.
-@param resize whether buf_pool_resize() is the caller */
-void btr_search_enable(bool resize)
+/** Enable the adaptive hash search system. */
+void btr_search_enable()
 {
-	if (!resize) {
-		buf_pool_mutex_enter_all();
-		if (srv_buf_pool_old_size != srv_buf_pool_size) {
-			buf_pool_mutex_exit_all();
-			return;
-		}
-		buf_pool_mutex_exit_all();
-	}
-
-	ulint hash_size = buf_pool_get_curr_size() / sizeof(void *) / 64;
-	btr_search_x_lock_all();
-
-	if (btr_search_sys->hash_tables) {
-		ut_ad(btr_search_enabled);
-		btr_search_x_unlock_all();
+	mutex_enter(&buf_pool.mutex);
+	if (srv_buf_pool_old_size != srv_buf_pool_size) {
+		mutex_exit(&buf_pool.mutex);
 		return;
 	}
+	mutex_exit(&buf_pool.mutex);
 
-	btr_search_sys->hash_tables = reinterpret_cast<hash_table_t**>(
-		ut_malloc(sizeof(hash_table_t*) * btr_ahi_parts, mem_key_ahi));
-	for (ulint i = 0; i < btr_ahi_parts; ++i) {
-		btr_search_sys->hash_tables[i] =
-			ib_create((hash_size / btr_ahi_parts),
-				  LATCH_ID_HASH_TABLE_MUTEX,
-				  0, MEM_HEAP_FOR_BTR_SEARCH);
-#if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
-                btr_search_sys->hash_tables[i]->adaptive = TRUE;
-#endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
-	}
-
+	btr_search_x_lock_all();
 	btr_search_enabled = true;
 	btr_search_x_unlock_all();
 }
 
+/** Returns the value of ref_count. The value is protected by latch.
+@param[in]	info		search info
+@param[in]	index		index identifier
+@return ref_count value. */
+ulint
+btr_search_info_get_ref_count(
+	btr_search_t*	info,
+	dict_index_t*	index)
+{
+	ulint ret = 0;
+
+	if (!btr_search_enabled) {
+		return(ret);
+	}
+
+	ut_ad(info);
+
+	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+	rw_lock_s_lock(ahi_latch);
+	ret = info->ref_count;
+	rw_lock_s_unlock(ahi_latch);
+
+	return(ret);
+}
+
 /** Updates the search info of an index about hash successes. NOTE that info
 is NOT protected by any semaphore, to save CPU time! Do not assume its fields
 are consistent.
@@ -468,7 +457,6 @@ btr_search_info_update_hash(
 	btr_cur_t*	cursor)
 {
 	dict_index_t*	index = cursor->index;
-	ulint		n_unique;
 	int		cmp;
 
 	ut_ad(!btr_search_own_any(RW_LOCK_S));
@@ -481,7 +469,7 @@ btr_search_info_update_hash(
 		return;
 	}
 
-	n_unique = dict_index_get_n_unique_in_tree(index);
+	uint16_t n_unique = dict_index_get_n_unique_in_tree(index);
 
 	if (info->n_hash_potential == 0) {
 
@@ -523,16 +511,13 @@ set_new_recomm:
 
 	cmp = ut_pair_cmp(cursor->up_match, cursor->up_bytes,
 			  cursor->low_match, cursor->low_bytes);
+	info->left_side = cmp >= 0;
+	info->n_hash_potential = cmp != 0;
+
 	if (cmp == 0) {
-		info->n_hash_potential = 0;
-
 		/* For extra safety, we set some sensible values here */
-
 		info->n_fields = 1;
 		info->n_bytes = 0;
-
-		info->left_side = TRUE;
-
 	} else if (cmp > 0) {
 		info->n_hash_potential = 1;
 
@@ -543,31 +528,31 @@ set_new_recomm:
 
 		} else if (cursor->low_match < cursor->up_match) {
 
-			info->n_fields = cursor->low_match + 1;
+			info->n_fields = static_cast<uint16_t>(
+				cursor->low_match + 1);
 			info->n_bytes = 0;
 		} else {
-			info->n_fields = cursor->low_match;
-			info->n_bytes = cursor->low_bytes + 1;
+			info->n_fields = static_cast<uint16_t>(
+				cursor->low_match);
+			info->n_bytes = static_cast<uint16_t>(
+				cursor->low_bytes + 1);
 		}
-
-		info->left_side = TRUE;
 	} else {
-		info->n_hash_potential = 1;
-
 		if (cursor->low_match >= n_unique) {
 
 			info->n_fields = n_unique;
 			info->n_bytes = 0;
 		} else if (cursor->low_match > cursor->up_match) {
 
-			info->n_fields = cursor->up_match + 1;
+			info->n_fields = static_cast<uint16_t>(
+				cursor->up_match + 1);
 			info->n_bytes = 0;
 		} else {
-			info->n_fields = cursor->up_match;
-			info->n_bytes = cursor->up_bytes + 1;
+			info->n_fields = static_cast<uint16_t>(
+				cursor->up_match);
+			info->n_bytes = static_cast<uint16_t>(
+				cursor->up_bytes + 1);
 		}
-
-		info->left_side = FALSE;
 	}
 }
 
@@ -652,48 +637,49 @@ btr_search_update_hash_ref(
 	buf_block_t*		block,
 	const btr_cur_t*	cursor)
 {
-	ut_ad(cursor->flag == BTR_CUR_HASH_FAIL);
+	dict_index_t*	index;
+	ulint		fold;
+	rec_t*		rec;
 
+	ut_ad(cursor->flag == BTR_CUR_HASH_FAIL);
+	ut_ad(rw_lock_own(btr_get_search_latch(cursor->index), RW_LOCK_X));
 	ut_ad(rw_lock_own_flagged(&block->lock,
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 	ut_ad(page_align(btr_cur_get_rec(cursor)) == block->frame);
 	ut_ad(page_is_leaf(block->frame));
 	assert_block_ahi_valid(block);
 
-	dict_index_t* index = block->index;
+	index = block->index;
+
+	if (!index) {
 
-	if (!index || !info->n_hash_potential) {
 		return;
 	}
 
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_ad(index == cursor->index);
 	ut_ad(!dict_index_is_ibuf(index));
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	rw_lock_x_lock(latch);
-	ut_ad(!block->index || block->index == index);
 
-	if (block->index
+	if ((info->n_hash_potential > 0)
 	    && (block->curr_n_fields == info->n_fields)
 	    && (block->curr_n_bytes == info->n_bytes)
-	    && (block->curr_left_side == info->left_side)
-	    && btr_search_enabled) {
+	    && (block->curr_left_side == info->left_side)) {
 		mem_heap_t*	heap		= NULL;
 		rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
 		rec_offs_init(offsets_);
 
-		const rec_t* rec = btr_cur_get_rec(cursor);
+		rec = btr_cur_get_rec(cursor);
 
 		if (!page_rec_is_user_rec(rec)) {
-			goto func_exit;
+
+			return;
 		}
 
-		ulint fold = rec_fold(
-			rec,
-			rec_get_offsets(rec, index, offsets_, true,
-					ULINT_UNDEFINED, &heap),
-			block->curr_n_fields,
-			block->curr_n_bytes, index->id);
+		fold = rec_fold(rec,
+				rec_get_offsets(rec, index, offsets_, true,
+						ULINT_UNDEFINED, &heap),
+				block->curr_n_fields,
+				block->curr_n_bytes, index->id);
 		if (UNIV_LIKELY_NULL(heap)) {
 			mem_heap_free(heap);
 		}
@@ -703,9 +689,6 @@ btr_search_update_hash_ref(
 
 		MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_ADDED);
 	}
-
-func_exit:
-	rw_lock_x_unlock(latch);
 }
 
 /** Checks if a guessed position for a tree cursor is right. Note that if
@@ -857,6 +840,80 @@ btr_search_failure(btr_search_t* info, btr_cur_t* cursor)
 	info->last_hash_succ = FALSE;
 }
 
+/** Clear the adaptive hash index on all pages in the buffer pool. */
+inline void buf_pool_t::clear_hash_index()
+{
+  ut_ad(btr_search_own_all(RW_LOCK_X));
+  ut_ad(!resizing);
+  ut_ad(!btr_search_enabled);
+
+  for (chunk_t *chunk= chunks + n_chunks; chunk-- != chunks; )
+  {
+    for (buf_block_t *block= chunk->blocks, * const end= block + chunk->size;
+         block != end; block++)
+    {
+      dict_index_t *index= block->index;
+      assert_block_ahi_valid(block);
+
+      /* We can clear block->index block->n_pointers when
+      btr_search_own_all(RW_LOCK_X); see the comments in buf0buf.h */
+      if (!index)
+      {
+# if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
+        ut_a(!block->n_pointers);
+# endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
+        continue;
+      }
+
+      ut_d(buf_page_state state= buf_block_get_state(block));
+      /* Another thread may have set the state to
+      BUF_BLOCK_REMOVE_HASH in buf_LRU_block_remove_hashed().
+
+      The state change in buf_pool_t::realloc() is not observable
+      here, because in that case we would have !block->index.
+
+      In the end, the entire adaptive hash index will be removed. */
+      ut_ad(state == BUF_BLOCK_FILE_PAGE || state == BUF_BLOCK_REMOVE_HASH);
+# if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
+      block->n_pointers= 0;
+# endif /* UNIV_AHI_DEBUG || UNIV_DEBUG */
+      block->index= nullptr;
+    }
+  }
+}
+
+/** Get a buffer block from an adaptive hash index pointer.
+This function does not return if the block is not identified.
+@param ptr  pointer to within a page frame
+@return pointer to block, never NULL */
+inline buf_block_t* buf_pool_t::block_from_ahi(const byte *ptr) const
+{
+  chunk_t::map *chunk_map = chunk_t::map_ref;
+  ut_ad(chunk_t::map_ref == chunk_t::map_reg);
+  ut_ad(!resizing);
+
+  chunk_t::map::const_iterator it= chunk_map->upper_bound(ptr);
+  ut_a(it != chunk_map->begin());
+
+  chunk_t *chunk= it == chunk_map->end()
+    ? chunk_map->rbegin()->second
+    : (--it)->second;
+
+  const size_t offs= size_t(ptr - chunk->blocks->frame) >> srv_page_size_shift;
+  ut_a(offs < chunk->size);
+
+  buf_block_t *block= &chunk->blocks[offs];
+  /* buf_pool_t::chunk_t::init() invokes buf_block_init() so that
+  block[n].frame == block->frame + n * srv_page_size.  Check it. */
+  ut_ad(block->frame == page_align(ptr));
+  /* Read the state of the block without holding a mutex.
+  A state transition from BUF_BLOCK_FILE_PAGE to
+  BUF_BLOCK_REMOVE_HASH is possible during this execution. */
+  ut_d(const buf_page_state state = buf_block_get_state(block));
+  ut_ad(state == BUF_BLOCK_FILE_PAGE || state == BUF_BLOCK_REMOVE_HASH);
+  return block;
+}
+
 /** Tries to guess the right search position based on the hash search info
 of the index. Note that if mode is PAGE_CUR_LE, which is used in inserts,
 and the function returns TRUE, then cursor->up_match and cursor->low_match
@@ -886,24 +943,24 @@ btr_search_guess_on_hash(
 	rw_lock_t*	ahi_latch,
 	mtr_t*		mtr)
 {
+	const rec_t*	rec;
 	ulint		fold;
 	index_id_t	index_id;
-#ifdef notdefined
-	btr_cur_t	cursor2;
-	btr_pcur_t	pcur;
-#endif
+
+	ut_ad(mtr->is_active());
 	ut_ad(!ahi_latch || rw_lock_own_flagged(
 		      ahi_latch, RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 
 	if (!btr_search_enabled) {
-		return false;
+		return(FALSE);
 	}
 
-	ut_ad(index && info && tuple && cursor && mtr);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_ad(!index->is_ibuf());
 	ut_ad(!ahi_latch || ahi_latch == btr_get_search_latch(index));
 	ut_ad((latch_mode == BTR_SEARCH_LEAF)
 	      || (latch_mode == BTR_MODIFY_LEAF));
+	compile_time_assert(ulint{BTR_SEARCH_LEAF} == ulint{RW_S_LATCH});
+	compile_time_assert(ulint{BTR_MODIFY_LEAF} == ulint{RW_X_LATCH});
 
 	/* Not supported for spatial index */
 	ut_ad(!dict_index_is_spatial(index));
@@ -912,14 +969,16 @@ btr_search_guess_on_hash(
 	any latch here! */
 
 	if (info->n_hash_potential == 0) {
-		return false;
+
+		return(FALSE);
 	}
 
 	cursor->n_fields = info->n_fields;
 	cursor->n_bytes = info->n_bytes;
 
 	if (dtuple_get_n_fields(tuple) < btr_search_get_n_fields(cursor)) {
-		return false;
+
+		return(FALSE);
 	}
 
 	index_id = index->id;
@@ -933,7 +992,6 @@ btr_search_guess_on_hash(
 	cursor->flag = BTR_CUR_HASH;
 
 	rw_lock_t* use_latch = ahi_latch ? NULL : btr_get_search_latch(index);
-	const rec_t* rec;
 
 	if (use_latch) {
 		rw_lock_s_lock(use_latch);
@@ -946,54 +1004,78 @@ btr_search_guess_on_hash(
 		ut_ad(rw_lock_own(ahi_latch, RW_LOCK_S));
 	}
 
-	rec = static_cast<const rec_t*>(
-		ha_search_and_get_data(btr_get_search_table(index), fold));
+	rec = (rec_t*) ha_search_and_get_data(
+		btr_get_search_table(index), fold);
 
-	if (!rec) {
+	if (rec == NULL) {
 		if (use_latch) {
 fail:
 			rw_lock_s_unlock(use_latch);
 		}
 
 		btr_search_failure(info, cursor);
-		return false;
+
+		return(FALSE);
 	}
 
-	buf_block_t*	block = buf_block_from_ahi(rec);
+	buf_block_t* block = buf_pool.block_from_ahi(rec);
 
 	if (use_latch) {
-		if (!buf_page_get_known_nowait(
-			latch_mode, block, BUF_MAKE_YOUNG,
-			__FILE__, __LINE__, mtr)) {
+		mutex_enter(&block->mutex);
+
+		if (buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH) {
+			/* Another thread is just freeing the block
+			from the LRU list of the buffer pool: do not
+			try to access this page. */
+			mutex_exit(&block->mutex);
 			goto fail;
 		}
 
-		const bool fail = index != block->index
-			&& index_id == block->index->id;
-		ut_a(!fail || block->index->freed());
+		ut_ad(buf_block_get_state(block) == BUF_BLOCK_FILE_PAGE);
+		DBUG_ASSERT(block->page.status != buf_page_t::FREED);
+
+		buf_page_set_accessed(&block->page);
+		buf_block_buf_fix_inc(block, __FILE__, __LINE__);
+		mutex_exit(&block->mutex);
+
+		buf_page_make_young_if_needed(&block->page);
+		mtr_memo_type_t	fix_type;
+		if (latch_mode == BTR_SEARCH_LEAF) {
+			if (!rw_lock_s_lock_nowait(&block->lock,
+						   __FILE__, __LINE__)) {
+got_no_latch:
+				buf_block_buf_fix_dec(block);
+				goto fail;
+			}
+			fix_type = MTR_MEMO_PAGE_S_FIX;
+		} else {
+			if (!rw_lock_x_lock_func_nowait_inline(
+				    &block->lock, __FILE__, __LINE__)) {
+				goto got_no_latch;
+			}
+			fix_type = MTR_MEMO_PAGE_X_FIX;
+		}
+		mtr->memo_push(block, fix_type);
+
+		buf_pool.stat.n_page_gets++;
+
 		rw_lock_s_unlock(use_latch);
 
 		buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
-		if (UNIV_UNLIKELY(fail)) {
-			goto fail_and_release_page;
-		}
-	} else if (UNIV_UNLIKELY(index != block->index
-				 && index_id == block->index->id)) {
-		ut_a(block->index->freed());
-		goto fail_and_release_page;
 	}
 
 	if (buf_block_get_state(block) != BUF_BLOCK_FILE_PAGE) {
 
 		ut_ad(buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH);
 
-fail_and_release_page:
 		if (!ahi_latch) {
+
 			btr_leaf_page_release(block, latch_mode, mtr);
 		}
 
 		btr_search_failure(info, cursor);
-		return false;
+
+		return(FALSE);
 	}
 
 	ut_ad(page_rec_is_user_rec(rec));
@@ -1009,7 +1091,14 @@ fail_and_release_page:
 	right. */
 	if (index_id != btr_page_get_index_id(block->frame)
 	    || !btr_search_check_guess(cursor, !!ahi_latch, tuple, mode)) {
-		goto fail_and_release_page;
+
+		if (!ahi_latch) {
+			btr_leaf_page_release(block, latch_mode, mtr);
+		}
+
+		btr_search_failure(info, cursor);
+
+		return(FALSE);
 	}
 
 	if (info->n_hash_potential < BTR_SEARCH_BUILD_LIMIT + 5) {
@@ -1057,17 +1146,12 @@ fail_and_release_page:
 #ifdef UNIV_SEARCH_PERF_STAT
 	btr_search_n_succ++;
 #endif
-	if (!ahi_latch && buf_page_peek_if_too_old(&block->page)) {
-
-		buf_page_make_young(&block->page);
-	}
-
 	/* Increment the page get statistics though we did not really
 	fix the page: for user info only */
-	{
-		buf_pool_t*	buf_pool = buf_pool_from_bpage(&block->page);
+	++buf_pool.stat.n_page_gets;
 
-		++buf_pool->stat.n_page_gets;
+	if (!ahi_latch) {
+		buf_page_make_young_if_needed(&block->page);
 	}
 
 	return true;
@@ -1077,7 +1161,7 @@ fail_and_release_page:
 @param[in,out]	block	block containing index page, s- or x-latched, or an
 			index page for which we know that
 			block->buf_fix_count == 0 or it is an index page which
-			has already been removed from the buf_pool->page_hash
+			has already been removed from the buf_pool.page_hash
 			i.e.: it is in state BUF_BLOCK_REMOVE_HASH */
 void btr_search_drop_page_hash_index(buf_block_t* block)
 {
@@ -1092,28 +1176,32 @@ void btr_search_drop_page_hash_index(buf_block_t* block)
 	ulint*			folds;
 	ulint			i;
 	mem_heap_t*		heap;
+	const dict_index_t*	index;
 	rec_offs*		offsets;
 	rw_lock_t*		latch;
+	btr_search_t*		info;
 
 retry:
+	/* Do a dirty check on block->index, return if the block is
+	not in the adaptive hash index. */
+	index = block->index;
 	/* This debug check uses a dirty read that could theoretically cause
-	false positives while buf_pool_clear_hash_index() is executing. */
+	false positives while buf_pool.clear_hash_index() is executing. */
 	assert_block_ahi_valid(block);
 	ut_ad(!btr_search_own_any(RW_LOCK_S));
 	ut_ad(!btr_search_own_any(RW_LOCK_X));
 
-	if (!block->index) {
+	if (index == NULL) {
 		return;
 	}
 
 	ut_ad(block->page.buf_fix_count == 0
 	      || buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH
 	      || rw_lock_own_flagged(&block->lock,
-				     RW_LOCK_FLAG_X | RW_LOCK_FLAG_S
-				     | RW_LOCK_FLAG_SX));
+				     RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 	ut_ad(page_is_leaf(block->frame));
 
-	/* We must not dereference block->index here, because it could be freed
+	/* We must not dereference index here, because it could be freed
 	if (index->table->n_ref_count == 0 && !mutex_own(&dict_sys.mutex)).
 	Determine the ahi_slot based on the block contents. */
 
@@ -1128,12 +1216,18 @@ retry:
 	rw_lock_s_lock(latch);
 	assert_block_ahi_valid(block);
 
-	if (!block->index || !btr_search_enabled) {
+	if (block->index == NULL) {
 		rw_lock_s_unlock(latch);
 		return;
 	}
 
-	dict_index_t* index = block->index;
+	/* The index associated with a block must remain the
+	same, because we are holding block->lock or the block is
+	not accessible by other threads (BUF_BLOCK_REMOVE_HASH),
+	or the index is not accessible to other threads
+	(buf_fix_count == 0 when DROP TABLE or similar is executing
+	buf_LRU_drop_page_hash_for_tablespace()). */
+	ut_a(index == block->index);
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	ut_ad(!index->disable_ahi);
 #endif
@@ -1141,7 +1235,27 @@ retry:
 
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index_id == index->id);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
+#ifdef UNIV_DEBUG
+	switch (dict_index_get_online_status(index)) {
+	case ONLINE_INDEX_CREATION:
+		/* The index is being created (bulk loaded). */
+	case ONLINE_INDEX_COMPLETE:
+		/* The index has been published. */
+	case ONLINE_INDEX_ABORTED:
+		/* Either the index creation was aborted due to an
+		error observed by InnoDB (in which case there should
+		not be any adaptive hash index entries), or it was
+		completed and then flagged aborted in
+		rollback_inplace_alter_table(). */
+		break;
+	case ONLINE_INDEX_ABORTED_DROPPED:
+		/* The index should have been dropped from the tablespace
+		already, and the adaptive hash index entries should have
+		been dropped as well. */
+		ut_error;
+	}
+#endif /* UNIV_DEBUG */
 
 	n_fields = block->curr_n_fields;
 	n_bytes = block->curr_n_bytes;
@@ -1229,14 +1343,9 @@ next_rec:
 			folds[i], page);
 	}
 
-	switch (index->search_info->ref_count--) {
-	case 0:
-		ut_error;
-	case 1:
-		if (index->freed()) {
-			btr_search_lazy_free(index);
-		}
-	}
+	info = btr_search_get_info(block->index);
+	ut_a(info->ref_count > 0);
+	info->ref_count--;
 
 	block->index = NULL;
 
@@ -1275,7 +1384,7 @@ void btr_search_drop_page_hash_when_freed(const page_id_t page_id)
 
 		/* If AHI is still valid, page can't be in free state.
 		AHI is dropped when page is freed. */
-		ut_ad(!block->page.file_page_was_freed);
+		DBUG_ASSERT(block->page.status != buf_page_t::FREED);
 
 		buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
 
@@ -1309,9 +1418,9 @@ btr_search_build_page_hash_index(
 	dict_index_t*	index,
 	buf_block_t*	block,
 	rw_lock_t*	ahi_latch,
-	ulint		n_fields,
-	ulint		n_bytes,
-	ibool		left_side)
+	uint16_t	n_fields,
+	uint16_t	n_bytes,
+	bool		left_side)
 {
 	const rec_t*	rec;
 	const rec_t*	next_rec;
@@ -1321,6 +1430,7 @@ btr_search_build_page_hash_index(
 	ulint		n_recs;
 	ulint*		folds;
 	const rec_t**	recs;
+	ulint		i;
 	mem_heap_t*	heap		= NULL;
 	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
 	rec_offs*	offsets		= offsets_;
@@ -1336,27 +1446,21 @@ btr_search_build_page_hash_index(
 	ut_ad(ahi_latch == btr_get_search_latch(index));
 	ut_ad(index);
 	ut_ad(block->page.id.space() == index->table->space_id);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
 	ut_ad(page_is_leaf(block->frame));
 
 	ut_ad(rw_lock_own_flagged(&block->lock,
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
-	ut_ad(block->page.id.page_no() >= 3);
 
 	rw_lock_s_lock(ahi_latch);
 
-	const bool enabled = btr_search_enabled;
-	const bool rebuild = enabled && block->index
+	const bool rebuild = block->index
 		&& (block->curr_n_fields != n_fields
 		    || block->curr_n_bytes != n_bytes
 		    || block->curr_left_side != left_side);
 
 	rw_lock_s_unlock(ahi_latch);
 
-	if (!enabled) {
-		return;
-	}
-
 	if (rebuild) {
 		btr_search_drop_page_hash_index(block);
 	}
@@ -1404,7 +1508,7 @@ btr_search_build_page_hash_index(
 		btr_search_get_n_fields(n_fields, n_bytes),
 		&heap);
 	ut_ad(page_rec_is_supremum(rec)
-	      || n_fields + (n_bytes > 0) == rec_offs_n_fields(offsets));
+	      || n_fields == rec_offs_n_fields(offsets) - (n_bytes > 0));
 
 	fold = rec_fold(rec, offsets, n_fields, n_bytes, index->id);
 
@@ -1457,12 +1561,19 @@ btr_search_build_page_hash_index(
 
 	btr_search_check_free_space_in_heap(index);
 
+	hash_table_t*	table	= btr_get_search_table(index);
 	rw_lock_x_lock(ahi_latch);
 
 	if (!btr_search_enabled) {
 		goto exit_func;
 	}
 
+	if (block->index && ((block->curr_n_fields != n_fields)
+			     || (block->curr_n_bytes != n_bytes)
+			     || (block->curr_left_side != left_side))) {
+		goto exit_func;
+	}
+
 	/* This counter is decremented every time we drop page
 	hash index entries and is incremented here. Since we can
 	rebuild hash index for a page that is already hashed, we
@@ -1471,24 +1582,18 @@ btr_search_build_page_hash_index(
 	if (!block->index) {
 		assert_block_ahi_empty(block);
 		index->search_info->ref_count++;
-	} else if (block->curr_n_fields != n_fields
-		   || block->curr_n_bytes != n_bytes
-		   || block->curr_left_side != left_side) {
-		goto exit_func;
 	}
 
 	block->n_hash_helps = 0;
 
-	block->curr_n_fields = unsigned(n_fields);
-	block->curr_n_bytes = unsigned(n_bytes);
-	block->curr_left_side = unsigned(left_side);
+	block->curr_n_fields = n_fields & dict_index_t::MAX_N_FIELDS;
+	block->curr_n_bytes = n_bytes & ((1U << 15) - 1);
+	block->curr_left_side = left_side;
 	block->index = index;
 
-	{
-		hash_table_t*	table = btr_get_search_table(index);
-		for (ulint i = 0; i < n_cached; i++) {
-			ha_insert_for_fold(table, folds[i], block, recs[i]);
-		}
+	for (i = 0; i < n_cached; i++) {
+
+		ha_insert_for_fold(table, folds[i], block, recs[i]);
 	}
 
 	MONITOR_INC(MONITOR_ADAPTIVE_HASH_PAGE_ADDED);
@@ -1538,7 +1643,11 @@ btr_search_info_update_slow(btr_search_t* info, btr_cur_t* cursor)
 		btr_search_n_hash_fail++;
 #endif /* UNIV_SEARCH_PERF_STAT */
 
+		rw_lock_x_lock(ahi_latch);
+
 		btr_search_update_hash_ref(info, block, cursor);
+
+		rw_lock_x_unlock(ahi_latch);
 	}
 
 	if (build_index) {
@@ -1594,9 +1703,9 @@ btr_search_move_or_delete_hash_entries(
 	rw_lock_s_lock(ahi_latch);
 
 	if (block->index) {
-		ulint	n_fields = block->curr_n_fields;
-		ulint	n_bytes = block->curr_n_bytes;
-		ibool	left_side = block->curr_left_side;
+		uint16_t n_fields = block->curr_n_fields;
+		uint16_t n_bytes = block->curr_n_bytes;
+		bool left_side = block->curr_left_side;
 
 		new_block->n_fields = block->curr_n_fields;
 		new_block->n_bytes = block->curr_n_bytes;
@@ -1623,6 +1732,7 @@ btr_search_move_or_delete_hash_entries(
 			using btr_cur_search_, the record is not yet deleted.*/
 void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 {
+	hash_table_t*	table;
 	buf_block_t*	block;
 	const rec_t*	rec;
 	ulint		fold;
@@ -1655,7 +1765,9 @@ void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index == cursor->index);
 	ut_a(block->curr_n_fields > 0 || block->curr_n_bytes > 0);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
+
+	table = btr_get_search_table(index);
 
 	rec = btr_cur_get_rec(cursor);
 
@@ -1666,24 +1778,22 @@ void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 		mem_heap_free(heap);
 	}
 
-	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+	rw_lock_t*	ahi_latch = btr_get_search_latch(index);
 
 	rw_lock_x_lock(ahi_latch);
 	assert_block_ahi_valid(block);
 
-	if (btr_search_enabled) {
-		hash_table_t* table = btr_get_search_table(index);
-		if (block->index) {
-			ut_a(block->index == index);
+	if (block->index) {
+		ut_a(block->index == index);
 
-			if (ha_search_and_delete_if_found(table, fold, rec)) {
-				MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVED);
-			} else {
-				MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVE_NOT_FOUND);
-			}
-
-			assert_block_ahi_valid(block);
+		if (ha_search_and_delete_if_found(table, fold, rec)) {
+			MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVED);
+		} else {
+			MONITOR_INC(
+				MONITOR_ADAPTIVE_HASH_ROW_REMOVE_NOT_FOUND);
 		}
+
+		assert_block_ahi_valid(block);
 	}
 
 	rw_lock_x_unlock(ahi_latch);
@@ -1726,10 +1836,10 @@ btr_search_update_hash_node_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	}
 
 	ut_a(cursor->index == index);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
 	rw_lock_x_lock(ahi_latch);
 
-	if (!block->index || !btr_search_enabled) {
+	if (!block->index) {
 
 		goto func_exit;
 	}
@@ -1779,6 +1889,8 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	ulint		next_fold = 0; /* remove warning (??? bug ???) */
 	ulint		n_fields;
 	ulint		n_bytes;
+	ibool		left_side;
+	bool		locked		= false;
 	mem_heap_t*	heap		= NULL;
 	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
 	rec_offs*	offsets		= offsets_;
@@ -1810,17 +1922,19 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	ut_ad(block->page.id.space() == index->table->space_id);
 	btr_search_check_free_space_in_heap(index);
 
+	table = btr_get_search_table(index);
+
 	rec = btr_cur_get_rec(cursor);
 
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	ut_a(!index->disable_ahi);
 #endif
 	ut_a(index == cursor->index);
-	ut_ad(!dict_index_is_ibuf(index));
+	ut_a(!dict_index_is_ibuf(index));
 
 	n_fields = block->curr_n_fields;
 	n_bytes = block->curr_n_bytes;
-	const bool left_side = block->curr_left_side;
+	left_side = block->curr_left_side;
 
 	ins_rec = page_rec_get_next_const(rec);
 	next_rec = page_rec_get_next_const(ins_rec);
@@ -1837,8 +1951,6 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 				     n_bytes, index->id);
 	}
 
-	bool locked = false;
-
 	if (!page_rec_is_infimum(rec) && !rec_is_metadata(rec, *index)) {
 		offsets = rec_get_offsets(
 			rec, index, offsets, true,
@@ -1849,11 +1961,10 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 			locked = true;
 			rw_lock_x_lock(ahi_latch);
 
-			if (!btr_search_enabled || !block->index) {
+			if (!btr_search_enabled) {
 				goto function_exit;
 			}
 
-			table = btr_get_search_table(index);
 			ha_insert_for_fold(table, ins_fold, block, ins_rec);
 		}
 
@@ -1866,10 +1977,9 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 			locked = true;
 			rw_lock_x_lock(ahi_latch);
 
-			if (!btr_search_enabled || !block->index) {
+			if (!btr_search_enabled) {
 				goto function_exit;
 			}
-			table = btr_get_search_table(index);
 		}
 
 		if (!left_side) {
@@ -1887,10 +1997,9 @@ check_next_rec:
 				locked = true;
 				rw_lock_x_lock(ahi_latch);
 
-				if (!btr_search_enabled || !block->index) {
+				if (!btr_search_enabled) {
 					goto function_exit;
 				}
-				table = btr_get_search_table(index);
 			}
 
 			ha_insert_for_fold(table, ins_fold, block, ins_rec);
@@ -1900,14 +2009,14 @@ check_next_rec:
 	}
 
 	if (ins_fold != next_fold) {
+
 		if (!locked) {
 			locked = true;
 			rw_lock_x_lock(ahi_latch);
 
-			if (!btr_search_enabled || !block->index) {
+			if (!btr_search_enabled) {
 				goto function_exit;
 			}
-			table = btr_get_search_table(index);
 		}
 
 		if (!left_side) {
@@ -1944,9 +2053,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
 	rec_offs*	offsets		= offsets_;
 
-	btr_search_x_lock_all();
 	if (!btr_search_enabled) {
-		btr_search_x_unlock_all();
 		return(TRUE);
 	}
 
@@ -1956,29 +2063,24 @@ btr_search_hash_table_validate(ulint hash_table_id)
 
 	rec_offs_init(offsets_);
 
-	buf_pool_mutex_enter_all();
+	btr_search_x_lock_all();
+	mutex_enter(&buf_pool.mutex);
 
 	cell_count = hash_get_n_cells(
-			btr_search_sys->hash_tables[hash_table_id]);
+		btr_search_sys->hash_tables[hash_table_id]);
 
 	for (i = 0; i < cell_count; i++) {
 		/* We release search latches every once in a while to
 		give other queries a chance to run. */
 		if ((i != 0) && ((i % chunk_size) == 0)) {
 
-			buf_pool_mutex_exit_all();
+			mutex_exit(&buf_pool.mutex);
 			btr_search_x_unlock_all();
 
 			os_thread_yield();
 
 			btr_search_x_lock_all();
-
-			if (!btr_search_enabled) {
-				ok = true;
-				goto func_exit;
-			}
-
-			buf_pool_mutex_enter_all();
+			mutex_enter(&buf_pool.mutex);
 
 			ulint	curr_cell_count = hash_get_n_cells(
 				btr_search_sys->hash_tables[hash_table_id]);
@@ -1998,13 +2100,10 @@ btr_search_hash_table_validate(ulint hash_table_id)
 
 		for (; node != NULL; node = node->next) {
 			const buf_block_t*	block
-				= buf_block_from_ahi((byte*) node->data);
+				= buf_pool.block_from_ahi((byte*) node->data);
 			const buf_block_t*	hash_block;
-			buf_pool_t*		buf_pool;
 			index_id_t		page_index_id;
 
-			buf_pool = buf_pool_from_bpage((buf_page_t*) block);
-
 			if (UNIV_LIKELY(buf_block_get_state(block)
 					== BUF_BLOCK_FILE_PAGE)) {
 
@@ -2014,7 +2113,6 @@ btr_search_hash_table_validate(ulint hash_table_id)
 				(BUF_BLOCK_REMOVE_HASH, see the
 				assertion and the comment below) */
 				hash_block = buf_block_hash_get(
-					buf_pool,
 					block->page.id);
 			} else {
 				hash_block = NULL;
@@ -2026,7 +2124,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 				/* When a block is being freed,
 				buf_LRU_search_and_free_block() first
 				removes the block from
-				buf_pool->page_hash by calling
+				buf_pool.page_hash by calling
 				buf_LRU_block_remove_hashed_page().
 				After that, it invokes
 				btr_search_drop_page_hash_index() to
@@ -2037,7 +2135,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 				     == BUF_BLOCK_REMOVE_HASH);
 			}
 
-			ut_ad(!dict_index_is_ibuf(block->index));
+			ut_a(!dict_index_is_ibuf(block->index));
 			ut_ad(block->page.id.space()
 			      == block->index->table->space_id);
 
@@ -2062,8 +2160,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 
 				ib::error() << "Error in an adaptive hash"
 					<< " index pointer to page "
-					<< page_id_t(page_get_space_id(page),
-						     page_get_page_no(page))
+					<< block->page.id
 					<< ", ptr mem address "
 					<< reinterpret_cast<const void*>(
 						node->data)
@@ -2089,20 +2186,13 @@ btr_search_hash_table_validate(ulint hash_table_id)
 		/* We release search latches every once in a while to
 		give other queries a chance to run. */
 		if (i != 0) {
-
-			buf_pool_mutex_exit_all();
+			mutex_exit(&buf_pool.mutex);
 			btr_search_x_unlock_all();
 
 			os_thread_yield();
 
 			btr_search_x_lock_all();
-
-			if (!btr_search_enabled) {
-				ok = true;
-				goto func_exit;
-			}
-
-			buf_pool_mutex_enter_all();
+			mutex_enter(&buf_pool.mutex);
 
 			ulint	curr_cell_count = hash_get_n_cells(
 				btr_search_sys->hash_tables[hash_table_id]);
@@ -2125,8 +2215,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 		}
 	}
 
-	buf_pool_mutex_exit_all();
-func_exit:
+	mutex_exit(&buf_pool.mutex);
 	btr_search_x_unlock_all();
 
 	if (UNIV_LIKELY_NULL(heap)) {
