@@ -41,12 +41,7 @@ Created 11/5/1995 Heikki Tuuri
 #include "srv0start.h"
 #include "srv0srv.h"
 
-/** There must be at least this many pages in buf_pool in the area to start
-a random read-ahead */
-#define BUF_READ_AHEAD_RANDOM_THRESHOLD(b)	\
-				(5 + BUF_READ_AHEAD_AREA(b) / 8)
-
-/** If there are buf_pool->curr_size per the number below pending reads, then
+/** If there are buf_pool.curr_size per the number below pending reads, then
 read-ahead is not done: this is to prevent flooding the buffer pool with
 i/o-fixed buffer blocks */
 #define BUF_READ_AHEAD_PEND_LIMIT	2
@@ -60,13 +55,12 @@ buf_read_page_handle_error(
 /*=======================*/
 	buf_page_t*	bpage)	/*!< in: pointer to the block */
 {
-	buf_pool_t*	buf_pool = buf_pool_from_bpage(bpage);
 	const bool	uncompressed = (buf_page_get_state(bpage)
 					== BUF_BLOCK_FILE_PAGE);
 	const page_id_t	old_page_id = bpage->id;
 
 	/* First unfix and release lock on the bpage */
-	buf_pool_mutex_enter(buf_pool);
+	mutex_enter(&buf_pool.mutex);
 	mutex_enter(buf_page_get_mutex(bpage));
 	ut_ad(buf_page_get_io_fix(bpage) == BUF_IO_READ);
 
@@ -85,10 +79,10 @@ buf_read_page_handle_error(
 	/* remove the block from LRU list */
 	buf_LRU_free_one_page(bpage, old_page_id);
 
-	ut_ad(buf_pool->n_pend_reads > 0);
-	buf_pool->n_pend_reads--;
+	ut_ad(buf_pool.n_pend_reads > 0);
+	buf_pool.n_pend_reads--;
 
-	buf_pool_mutex_exit(buf_pool);
+	mutex_exit(&buf_pool.mutex);
 }
 
 /** Low-level function which reads a page asynchronously from a file to the
@@ -100,12 +94,11 @@ flag is cleared and the x-lock released by an i/o-handler thread.
 			if we are trying
 			to read from a non-existent tablespace
 @param[in] sync		true if synchronous aio is desired
-@param[in] type		IO type, SIMULATED, IGNORE_MISSING
 @param[in] mode		BUF_READ_IBUF_PAGES_ONLY, ...,
 @param[in] page_id	page id
 @param[in] zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in] unzip	true=request uncompressed page
-@param[in] ignore_missing_space  true=ignore missing space when reading
+@param[in] ignore	whether to ignore out-of-bounds page_id
 @return 1 if a read request was queued, 0 if the page already resided
 in buf_pool, or if the page is in the doublewrite buffer blocks in
 which case it is never read into the pool, or if the tablespace does
@@ -115,12 +108,11 @@ ulint
 buf_read_page_low(
 	dberr_t*		err,
 	bool			sync,
-	ulint			type,
 	ulint			mode,
 	const page_id_t		page_id,
 	ulint			zip_size,
 	bool			unzip,
-	bool			ignore_missing_space = false)
+	bool			ignore = false)
 {
 	buf_page_t*	bpage;
 
@@ -176,20 +168,17 @@ buf_read_page_low(
 		dst = ((buf_block_t*) bpage)->frame;
 	}
 
-	IORequest	request(type | IORequest::READ);
-
 	*err = fil_io(
-		request, sync, page_id, zip_size, 0,
+		IORequestRead, sync, page_id, zip_size, 0,
 		zip_size ? zip_size : srv_page_size,
-		dst, bpage, ignore_missing_space);
+		dst, bpage, ignore);
 
 	if (sync) {
 		thd_wait_end(NULL);
 	}
 
 	if (UNIV_UNLIKELY(*err != DB_SUCCESS)) {
-		if (IORequest::ignore_missing(type)
-		    || *err == DB_TABLESPACE_DELETED) {
+		if (ignore || *err == DB_TABLESPACE_DELETED) {
 			buf_read_page_handle_error(bpage);
 			return(0);
 		}
@@ -229,15 +218,12 @@ get read even if we return a positive value! */
 ulint
 buf_read_ahead_random(const page_id_t page_id, ulint zip_size, bool ibuf)
 {
-	buf_pool_t*	buf_pool = buf_pool_get(page_id);
 	ulint		recent_blocks	= 0;
 	ulint		ibuf_mode;
 	ulint		count;
 	ulint		low, high;
 	dberr_t		err = DB_SUCCESS;
 	ulint		i;
-	const ulint	buf_read_ahead_random_area
-				= BUF_READ_AHEAD_AREA(buf_pool);
 
 	if (!srv_random_read_ahead) {
 		/* Disabled by user */
@@ -258,6 +244,8 @@ buf_read_ahead_random(const page_id_t page_id, ulint zip_size, bool ibuf)
 		return(0);
 	}
 
+	const ulint	buf_read_ahead_random_area
+		= buf_pool.read_ahead_area;
 	low  = (page_id.page_no() / buf_read_ahead_random_area)
 		* buf_read_ahead_random_area;
 
@@ -294,11 +282,11 @@ buf_read_ahead_random(const page_id_t page_id, ulint zip_size, bool ibuf)
 		return(0);
 	}
 
-	buf_pool_mutex_enter(buf_pool);
+	mutex_enter(&buf_pool.mutex);
 
-	if (buf_pool->n_pend_reads
-	    > buf_pool->curr_size / BUF_READ_AHEAD_PEND_LIMIT) {
-		buf_pool_mutex_exit(buf_pool);
+	if (buf_pool.n_pend_reads
+	    > buf_pool.curr_size / BUF_READ_AHEAD_PEND_LIMIT) {
+		mutex_exit(&buf_pool.mutex);
 
 		return(0);
 	}
@@ -307,25 +295,19 @@ buf_read_ahead_random(const page_id_t page_id, ulint zip_size, bool ibuf)
 	that is, reside near the start of the LRU list. */
 
 	for (i = low; i < high; i++) {
-		const buf_page_t*	bpage = buf_page_hash_get(
-			buf_pool, page_id_t(page_id.space(), i));
-
-		if (bpage != NULL
-		    && buf_page_is_accessed(bpage)
-		    && buf_page_peek_if_young(bpage)) {
-
-			recent_blocks++;
-
-			if (recent_blocks
-			    >= BUF_READ_AHEAD_RANDOM_THRESHOLD(buf_pool)) {
-
-				buf_pool_mutex_exit(buf_pool);
+		if (const buf_page_t* bpage = buf_page_hash_get(
+			    page_id_t(page_id.space(), i))) {
+			if (buf_page_is_accessed(bpage)
+			    && buf_page_peek_if_young(bpage)
+			    && ++recent_blocks
+			    >= 5 + buf_pool.read_ahead_area / 8) {
+				mutex_exit(&buf_pool.mutex);
 				goto read_ahead;
 			}
 		}
 	}
 
-	buf_pool_mutex_exit(buf_pool);
+	mutex_exit(&buf_pool.mutex);
 	/* Do nothing */
 	return(0);
 
@@ -344,7 +326,6 @@ read_ahead:
 		if (!ibuf_bitmap_page(cur_page_id, zip_size)) {
 			count += buf_read_page_low(
 				&err, false,
-				IORequest::DO_NOT_WAKE,
 				ibuf_mode,
 				cur_page_id, zip_size, false);
 
@@ -364,11 +345,6 @@ read_ahead:
 		}
 	}
 
-	/* In simulated aio we wake the aio handler threads only after
-	queuing all aio requests, in native aio the following call does
-	nothing: */
-
-	os_aio_simulated_wake_handler_threads();
 
 	if (count) {
 		DBUG_PRINT("ib_buf", ("random read-ahead %u pages, %u:%u",
@@ -381,7 +357,7 @@ read_ahead:
 	LRU policy decision. */
 	buf_LRU_stat_inc_io();
 
-	buf_pool->stat.n_ra_pages_read_rnd += count;
+	buf_pool.stat.n_ra_pages_read_rnd += count;
 	srv_stats.buf_pool_reads.add(count);
 	return(count);
 }
@@ -409,8 +385,7 @@ dberr_t buf_read_page(const page_id_t page_id, ulint zip_size)
 	of the buffer pool mutex becomes an expensive bottleneck. */
 
 	count = buf_read_page_low(
-		&err, true,
-		0, BUF_READ_ANY_PAGE, page_id, zip_size, false);
+		&err, true, BUF_READ_ANY_PAGE, page_id, zip_size, false);
 
 	srv_stats.buf_pool_reads.add(count);
 
@@ -440,9 +415,8 @@ buf_read_page_background(const page_id_t page_id, ulint zip_size, bool sync)
 
 	count = buf_read_page_low(
 		&err, sync,
-		IORequest::DO_NOT_WAKE | IORequest::IGNORE_MISSING,
 		BUF_READ_ANY_PAGE,
-		page_id, zip_size, false);
+		page_id, zip_size, false, true);
 
 	switch (err) {
 	case DB_SUCCESS:
@@ -503,7 +477,6 @@ which could result in a deadlock if the OS does not support asynchronous io.
 ulint
 buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 {
-	buf_pool_t*	buf_pool = buf_pool_get(page_id);
 	buf_page_t*	bpage;
 	buf_frame_t*	frame;
 	buf_page_t*	pred_bpage	= NULL;
@@ -515,8 +488,6 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 	ulint		low, high;
 	dberr_t		err = DB_SUCCESS;
 	ulint		i;
-	const ulint	buf_read_ahead_linear_area
-		= BUF_READ_AHEAD_AREA(buf_pool);
 	ulint		threshold;
 
 	/* check if readahead is disabled */
@@ -529,6 +500,8 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 		return(0);
 	}
 
+	const ulint	buf_read_ahead_linear_area
+		= buf_pool.read_ahead_area;
 	low  = (page_id.page_no() / buf_read_ahead_linear_area)
 		* buf_read_ahead_linear_area;
 	high = (page_id.page_no() / buf_read_ahead_linear_area + 1)
@@ -566,11 +539,11 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 		return(0);
 	}
 
-	buf_pool_mutex_enter(buf_pool);
+	mutex_enter(&buf_pool.mutex);
 
-	if (buf_pool->n_pend_reads
-	    > buf_pool->curr_size / BUF_READ_AHEAD_PEND_LIMIT) {
-		buf_pool_mutex_exit(buf_pool);
+	if (buf_pool.n_pend_reads
+	    > buf_pool.curr_size / BUF_READ_AHEAD_PEND_LIMIT) {
+		mutex_exit(&buf_pool.mutex);
 
 		return(0);
 	}
@@ -588,13 +561,12 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 	/* How many out of order accessed pages can we ignore
 	when working out the access pattern for linear readahead */
 	threshold = ut_min(static_cast<ulint>(64 - srv_read_ahead_threshold),
-			   BUF_READ_AHEAD_AREA(buf_pool));
+			   buf_pool.read_ahead_area);
 
 	fail_count = 0;
 
 	for (i = low; i < high; i++) {
-		bpage = buf_page_hash_get(buf_pool,
-					  page_id_t(page_id.space(), i));
+		bpage = buf_page_hash_get(page_id_t(page_id.space(), i));
 
 		if (bpage == NULL || !buf_page_is_accessed(bpage)) {
 			/* Not accessed */
@@ -620,7 +592,7 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 
 		if (fail_count > threshold) {
 			/* Too many failures: return */
-			buf_pool_mutex_exit(buf_pool);
+			mutex_exit(&buf_pool.mutex);
 			return(0);
 		}
 
@@ -632,10 +604,10 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 	/* If we got this far, we know that enough pages in the area have
 	been accessed in the right order: linear read-ahead can be sensible */
 
-	bpage = buf_page_hash_get(buf_pool, page_id);
+	bpage = buf_page_hash_get(page_id);
 
 	if (bpage == NULL) {
-		buf_pool_mutex_exit(buf_pool);
+		mutex_exit(&buf_pool.mutex);
 
 		return(0);
 	}
@@ -658,10 +630,11 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 	prevent deadlocks. Even if we read values which are nonsense, the
 	algorithm will work. */
 
-	pred_offset = fil_page_get_prev(frame);
-	succ_offset = fil_page_get_next(frame);
-
-	buf_pool_mutex_exit(buf_pool);
+	pred_offset = mach_read_from_4(my_assume_aligned<4>(FIL_PAGE_PREV
+							    + frame));
+	succ_offset = mach_read_from_4(my_assume_aligned<4>(FIL_PAGE_NEXT
+							    + frame));
+	mutex_exit(&buf_pool.mutex);
 
 	if ((page_id.page_no() == low)
 	    && (succ_offset == page_id.page_no() + 1)) {
@@ -703,12 +676,6 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 
 	ulint ibuf_mode = ibuf ? BUF_READ_IBUF_PAGES_ONLY : BUF_READ_ANY_PAGE;
 
-	/* Since Windows XP seems to schedule the i/o handler thread
-	very eagerly, and consequently it does not wait for the
-	full read batch to be posted, we use special heuristics here */
-
-	os_aio_simulated_put_read_threads_to_sleep();
-
 	for (i = low; i < high; i++) {
 		/* It is only sensible to do read-ahead in the non-sync
 		aio mode: hence FALSE as the first parameter */
@@ -718,7 +685,6 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 		if (!ibuf_bitmap_page(cur_page_id, zip_size)) {
 			count += buf_read_page_low(
 				&err, false,
-				IORequest::DO_NOT_WAKE,
 				ibuf_mode, cur_page_id, zip_size, false);
 
 			switch (err) {
@@ -738,12 +704,6 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 		}
 	}
 
-	/* In simulated aio we wake the aio handler threads only after
-	queuing all aio requests, in native aio the following call does
-	nothing: */
-
-	os_aio_simulated_wake_handler_threads();
-
 	if (count) {
 		DBUG_PRINT("ib_buf", ("linear read-ahead " ULINTPF " pages, "
 				      "%u:%u",
@@ -756,109 +716,10 @@ buf_read_ahead_linear(const page_id_t page_id, ulint zip_size, bool ibuf)
 	LRU policy decision. */
 	buf_LRU_stat_inc_io();
 
-	buf_pool->stat.n_ra_pages_read += count;
+	buf_pool.stat.n_ra_pages_read += count;
 	return(count);
 }
 
-/********************************************************************//**
-Issues read requests for pages which the ibuf module wants to read in, in
-order to contract the insert buffer tree. Technically, this function is like
-a read-ahead function. */
-void
-buf_read_ibuf_merge_pages(
-/*======================*/
-	bool		sync,		/*!< in: true if the caller
-					wants this function to wait
-					for the highest address page
-					to get read in, before this
-					function returns */
-	const ulint*	space_ids,	/*!< in: array of space ids */
-	const ulint*	page_nos,	/*!< in: array of page numbers
-					to read, with the highest page
-					number the last in the
-					array */
-	ulint		n_stored)	/*!< in: number of elements
-					in the arrays */
-{
-#ifdef UNIV_IBUF_DEBUG
-	ut_a(n_stored < srv_page_size);
-#endif
-
-	for (ulint i = 0; i < n_stored; i++) {
-		fil_space_t* space = fil_space_acquire_silent(space_ids[i]);
-		if (!space) {
-tablespace_deleted:
-			/* The tablespace was not found: remove all
-			entries for it */
-			ibuf_delete_for_discarded_space(space_ids[i]);
-			while (i + 1 < n_stored
-			       && space_ids[i + 1] == space_ids[i]) {
-				i++;
-			}
-			continue;
-		}
-
-		if (UNIV_UNLIKELY(page_nos[i] >= space->size)) {
-			do {
-				ibuf_delete_recs(page_id_t(space_ids[i],
-							   page_nos[i]));
-			} while (++i < n_stored
-				 && space_ids[i - 1] == space_ids[i]
-				 && page_nos[i] >= space->size);
-			i--;
-next:
-			space->release();
-			continue;
-		}
-
-		const page_id_t	page_id(space_ids[i], page_nos[i]);
-
-		buf_pool_t*	buf_pool = buf_pool_get(page_id);
-
-		while (buf_pool->n_pend_reads
-		       > buf_pool->curr_size / BUF_READ_AHEAD_PEND_LIMIT) {
-			os_thread_sleep(500000);
-		}
-
-		dberr_t	err;
-
-		buf_read_page_low(&err,
-				  sync && (i + 1 == n_stored),
-				  0,
-				  BUF_READ_ANY_PAGE, page_id,
-				  space->zip_size(),
-				  true, true /* ignore_missing_space */);
-
-		switch(err) {
-		case DB_SUCCESS:
-		case DB_ERROR:
-			break;
-		case DB_TABLESPACE_DELETED:
-			space->release();
-			goto tablespace_deleted;
-		case DB_PAGE_CORRUPTED:
-		case DB_DECRYPTION_FAILED:
-			ib::error() << "Failed to read or decrypt page "
-				    << page_nos[i]
-				    << " of '" << space->chain.start->name
-				    << "' for change buffer merge";
-			break;
-		default:
-			ut_error;
-		}
-
-		goto next;
-	}
-
-	os_aio_simulated_wake_handler_threads();
-
-	if (n_stored) {
-		DBUG_PRINT("ib_buf",
-			   ("ibuf merge read-ahead %u pages, space %u",
-			    unsigned(n_stored), unsigned(space_ids[0])));
-	}
-}
-
 /** Issues read requests for pages which recovery wants to read in.
 @param[in]	sync		true if the caller wants this function to wait
 for the highest address page to get read in, before this function returns
@@ -885,48 +746,30 @@ buf_read_recv_pages(
 	const ulint zip_size = space->zip_size();
 
 	for (ulint i = 0; i < n_stored; i++) {
-		buf_pool_t*		buf_pool;
 		const page_id_t	cur_page_id(space_id, page_nos[i]);
 
-		ulint			count = 0;
-
-		buf_pool = buf_pool_get(cur_page_id);
 		ulint limit = 0;
-		for (ulint j = 0; j < buf_pool->n_chunks; j++) {
-			limit += buf_pool->chunks[j].size / 2;
+		for (ulint j = 0; j < buf_pool.n_chunks; j++) {
+			limit += buf_pool.chunks[j].size / 2;
 		}
 
-		while (buf_pool->n_pend_reads >= limit) {
-			os_aio_simulated_wake_handler_threads();
+		for (ulint count = 0; buf_pool.n_pend_reads >= limit; ) {
 			os_thread_sleep(10000);
 
-			count++;
-
-			if (!(count % 1000)) {
+			if (!(++count % 1000)) {
 
 				ib::error()
 					<< "Waited for " << count / 100
 					<< " seconds for "
-					<< buf_pool->n_pend_reads
+					<< buf_pool.n_pend_reads
 					<< " pending reads";
 			}
 		}
 
 		dberr_t err;
-
-		if (sync && i + 1 == n_stored) {
-			buf_read_page_low(
-				&err, true,
-				0,
-				BUF_READ_ANY_PAGE,
-				cur_page_id, zip_size, true);
-		} else {
-			buf_read_page_low(
-				&err, false,
-				IORequest::DO_NOT_WAKE,
-				BUF_READ_ANY_PAGE,
-				cur_page_id, zip_size, true);
-		}
+		buf_read_page_low(
+			&err, sync && i + 1 == n_stored,
+			BUF_READ_ANY_PAGE, cur_page_id, zip_size, true);
 
 		if (err == DB_DECRYPTION_FAILED || err == DB_PAGE_CORRUPTED) {
 			ib::error() << "Recovery failed to read or decrypt "
@@ -934,8 +777,6 @@ buf_read_recv_pages(
 		}
 	}
 
-	os_aio_simulated_wake_handler_threads();
-
 	DBUG_PRINT("ib_buf", ("recovery read-ahead (%u pages)",
 			      unsigned(n_stored)));
 }
