@@ -111,6 +111,9 @@ row_undo_mod_clust_low(
 	ut_ad(rec_get_trx_id(btr_cur_get_rec(btr_cur),
 			     btr_cur_get_index(btr_cur))
 	      == thr_get_trx(thr)->id);
+	ut_ad(node->ref != &trx_undo_metadata
+	      || node->update->info_bits == REC_INFO_METADATA_ADD
+	      || node->update->info_bits == REC_INFO_METADATA_ALTER);
 
 	if (mode != BTR_MODIFY_LEAF
 	    && dict_index_is_online_ddl(btr_cur_get_index(btr_cur))) {
@@ -131,6 +134,7 @@ row_undo_mod_clust_low(
 			btr_cur, offsets, offsets_heap,
 			node->update, node->cmpl_info,
 			thr, thr_get_trx(thr)->id, mtr);
+		ut_ad(err != DB_SUCCESS || node->ref != &trx_undo_metadata);
 	} else {
 		big_rec_t*	dummy_big_rec;
 
@@ -143,6 +147,28 @@ row_undo_mod_clust_low(
 			node->cmpl_info, thr, thr_get_trx(thr)->id, mtr);
 
 		ut_a(!dummy_big_rec);
+
+		if (err == DB_SUCCESS
+		    && node->ref == &trx_undo_metadata
+		    && btr_cur_get_index(btr_cur)->table->instant
+		    && node->update->info_bits == REC_INFO_METADATA_ADD) {
+			btr_reset_instant(*btr_cur_get_index(btr_cur), false,
+					  mtr);
+		}
+	}
+
+	if (err == DB_SUCCESS
+	    && btr_cur_get_index(btr_cur)->table->id == DICT_COLUMNS_ID) {
+		/* This is rolling back an UPDATE or DELETE on SYS_COLUMNS.
+		If it was part of an instant ALTER TABLE operation, we
+		must evict the table definition, so that it can be
+		reloaded after the dictionary operation has been
+		completed. At this point, any corresponding operation
+		to the metadata record will have been rolled back. */
+		const dfield_t& table_id = *dtuple_get_nth_field(node->row, 0);
+		ut_ad(dfield_get_len(&table_id) == 8);
+		node->trx->evict_table(mach_read_from_8(static_cast<byte*>(
+					table_id.data)));
 	}
 
 	return(err);
@@ -221,7 +247,7 @@ row_undo_mod_clust(
 	ut_ad(thr_get_trx(thr) == node->trx);
 	ut_ad(node->trx->dict_operation_lock_mode);
 	ut_ad(node->trx->in_rollback);
-	ut_ad(rw_lock_own_flagged(&dict_operation_lock,
+	ut_ad(rw_lock_own_flagged(&dict_sys.latch,
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 
 	log_free_check();
@@ -278,7 +304,7 @@ row_undo_mod_clust(
 	}
 
 	/* Online rebuild cannot be initiated while we are holding
-	dict_operation_lock and index->lock. (It can be aborted.) */
+	dict_sys.latch and index->lock. (It can be aborted.) */
 	ut_ad(online || !dict_index_is_online_ddl(index));
 
 	if (err == DB_SUCCESS && online) {
@@ -401,33 +427,65 @@ row_undo_mod_clust(
 			goto mtr_commit_exit;
 		}
 
+		ulint trx_id_offset = index->trx_id_offset;
 		ulint trx_id_pos = index->n_uniq ? index->n_uniq : 1;
-		ut_ad(index->n_uniq <= MAX_REF_PARTS);
-		/* Reserve enough offsets for the PRIMARY KEY and 2 columns
-		so that we can access DB_TRX_ID, DB_ROLL_PTR. */
+		/* Reserve enough offsets for the PRIMARY KEY and
+		2 columns so that we can access DB_TRX_ID, DB_ROLL_PTR. */
 		rec_offs offsets_[REC_OFFS_HEADER_SIZE + MAX_REF_PARTS + 2];
-		rec_offs_init(offsets_);
-		offsets = rec_get_offsets(
-			rec, index, offsets_, true, trx_id_pos + 2, &heap);
-		ulint len;
-		ulint trx_id_offset = rec_get_nth_field_offs(
-			offsets, trx_id_pos, &len);
-		ut_ad(len == DATA_TRX_ID_LEN);
+		if (trx_id_offset) {
+#ifdef UNIV_DEBUG
+			ut_ad(rec_offs_validate(NULL, index, offsets));
+			if (buf_block_get_page_zip(
+				    btr_pcur_get_block(&node->pcur))) {
+				/* Below, page_zip_write_trx_id_and_roll_ptr()
+				needs offsets to access DB_TRX_ID,DB_ROLL_PTR.
+				We already computed offsets for possibly
+				another record in the clustered index.
+				Because the PRIMARY KEY is fixed-length,
+				the offsets for the PRIMARY KEY and
+				DB_TRX_ID,DB_ROLL_PTR are still valid.
+				Silence the rec_offs_validate() assertion. */
+				rec_offs_make_valid(rec, index, true, offsets);
+			}
+#endif
+		} else if (rec_is_metadata(rec, *index)) {
+			ut_ad(!buf_block_get_page_zip(btr_pcur_get_block(
+							      pcur)));
+			for (unsigned i = index->first_user_field(); i--; ) {
+				trx_id_offset += index->fields[i].fixed_len;
+			}
+		} else {
+			ut_ad(index->n_uniq <= MAX_REF_PARTS);
+			rec_offs_init(offsets_);
+			offsets = rec_get_offsets(
+				rec, index, offsets_, true, trx_id_pos + 2,
+				&heap);
+			ulint len;
+			trx_id_offset = rec_get_nth_field_offs(
+				offsets, trx_id_pos, &len);
+			ut_ad(len == DATA_TRX_ID_LEN);
+		}
 
 		if (trx_read_trx_id(rec + trx_id_offset) == node->new_trx_id) {
 			ut_ad(!rec_get_deleted_flag(
-				      rec, dict_table_is_comp(node->table)));
+				      rec, dict_table_is_comp(node->table))
+			      || rec_is_alter_metadata(rec, *index));
 			index->set_modified(mtr);
-			if (page_zip_des_t* page_zip = buf_block_get_page_zip(
-				    btr_pcur_get_block(&node->pcur))) {
+			buf_block_t* block = btr_pcur_get_block(pcur);
+			if (UNIV_LIKELY_NULL(block->page.zip.data)) {
 				page_zip_write_trx_id_and_roll_ptr(
-					page_zip, rec, offsets, trx_id_pos,
+					block, rec, offsets, trx_id_pos,
 					0, 1ULL << ROLL_PTR_INSERT_FLAG_POS,
 					&mtr);
 			} else {
-				mlog_write_string(rec + trx_id_offset,
-						  reset_trx_id,
-						  sizeof reset_trx_id, &mtr);
+				size_t offs = page_offset(rec + trx_id_offset);
+				mtr.memset(block, offs, DATA_TRX_ID_LEN, 0);
+				offs += DATA_TRX_ID_LEN;
+				mtr.write<1,mtr_t::MAYBE_NOP>(*block,
+							      block->frame
+							      + offs, 0x80U);
+				mtr.memset(block, offs + 1,
+					   DATA_ROLL_PTR_LEN - 1, 0);
 			}
 		}
 	} else {
@@ -438,8 +496,6 @@ mtr_commit_exit:
 	btr_pcur_commit_specify_mtr(pcur, &mtr);
 
 func_exit:
-	node->state = UNDO_NODE_FETCH_NEXT;
-
 	if (offsets_heap) {
 		mem_heap_free(offsets_heap);
 	}
@@ -545,9 +601,8 @@ row_undo_mod_del_mark_or_remove_sec_low(
 	    || row_vers_old_has_index_entry(
 		    false, btr_pcur_get_rec(&node->pcur),
 		    &mtr_vers, index, entry, 0, 0)) {
-		err = btr_cur_del_mark_set_sec_rec(BTR_NO_LOCKING_FLAG,
-						   btr_cur, TRUE, thr, &mtr);
-		ut_ad(err == DB_SUCCESS);
+		btr_rec_set_deleted<true>(btr_cur_get_block(btr_cur),
+					  btr_cur_get_rec(btr_cur), &mtr);
 	} else {
 		/* Remove the index record */
 
@@ -782,11 +837,8 @@ try_again:
 
 		break;
 	case ROW_FOUND:
-		err = btr_cur_del_mark_set_sec_rec(
-			BTR_NO_LOCKING_FLAG,
-			btr_cur, FALSE, thr, &mtr);
-
-		ut_a(err == DB_SUCCESS);
+		btr_rec_set_deleted<false>(btr_cur_get_block(btr_cur),
+					   btr_cur_get_rec(btr_cur), &mtr);
 		heap = mem_heap_create(
 			sizeof(upd_t)
 			+ dtuple_get_n_fields(entry) * sizeof(upd_field_t));
@@ -852,9 +904,9 @@ row_undo_mod_sec_flag_corrupted(
 		on the data dictionary during normal rollback,
 		we can only mark the index corrupted in the
 		data dictionary cache. TODO: fix this somehow.*/
-		mutex_enter(&dict_sys->mutex);
+		mutex_enter(&dict_sys.mutex);
 		dict_set_corrupted_index_cache_only(index);
-		mutex_exit(&dict_sys->mutex);
+		mutex_exit(&dict_sys.mutex);
 		break;
 	default:
 		ut_ad(0);
@@ -913,7 +965,7 @@ row_undo_mod_upd_del_sec(
 			does not exist.  However, this situation may
 			only occur during the rollback of incomplete
 			transactions. */
-			ut_a(thr_get_trx(thr) == trx_roll_crash_recv_trx);
+			ut_a(thr_is_recv(thr));
 		} else {
 			err = row_undo_mod_del_mark_or_remove_sec(
 				node, thr, index, entry);
@@ -1141,14 +1193,10 @@ row_undo_mod_upd_exist_sec(
 	return(err);
 }
 
-/***********************************************************//**
-Parses the row reference and other info in a modify undo log record. */
-static MY_ATTRIBUTE((nonnull))
-void
-row_undo_mod_parse_undo_rec(
-/*========================*/
-	undo_node_t*	node,		/*!< in: row undo node */
-	ibool		dict_locked)	/*!< in: TRUE if own dict_sys->mutex */
+/** Parse an update undo record.
+@param[in,out]	node		row rollback state
+@param[in]	dict_locked	whether the data dictionary cache is locked */
+static bool row_undo_mod_parse_undo_rec(undo_node_t* node, bool dict_locked)
 {
 	dict_index_t*	clust_index;
 	byte*		ptr;
@@ -1156,24 +1204,33 @@ row_undo_mod_parse_undo_rec(
 	table_id_t	table_id;
 	trx_id_t	trx_id;
 	roll_ptr_t	roll_ptr;
-	ulint		info_bits;
+	byte		info_bits;
 	ulint		type;
 	ulint		cmpl_info;
 	bool		dummy_extern;
 
+	ut_ad(node->state == UNDO_UPDATE_PERSISTENT
+	      || node->state == UNDO_UPDATE_TEMPORARY);
+	ut_ad(node->trx->in_rollback);
+	ut_ad(!trx_undo_roll_ptr_is_insert(node->roll_ptr));
+
 	ptr = trx_undo_rec_get_pars(node->undo_rec, &type, &cmpl_info,
 				    &dummy_extern, &undo_no, &table_id);
 	node->rec_type = type;
 
-	node->table = dict_table_open_on_id(
-		table_id, dict_locked, DICT_TABLE_OP_NORMAL);
+	if (node->state == UNDO_UPDATE_PERSISTENT) {
+		node->table = dict_table_open_on_id(table_id, dict_locked,
+						    DICT_TABLE_OP_NORMAL);
+	} else if (!dict_locked) {
+		mutex_enter(&dict_sys.mutex);
+		node->table = dict_sys.get_temporary_table(table_id);
+		mutex_exit(&dict_sys.mutex);
+	} else {
+		node->table = dict_sys.get_temporary_table(table_id);
+	}
 
-	/* TODO: other fixes associated with DROP TABLE + rollback in the
-	same table by another user */
-
-	if (node->table == NULL) {
-		/* Table was dropped */
-		return;
+	if (!node->table) {
+		return false;
 	}
 
 	ut_ad(!node->table->skip_alter_undo);
@@ -1191,7 +1248,7 @@ close_table:
 		connection, instead of doing this rollback. */
 		dict_table_close(node->table, dict_locked, FALSE);
 		node->table = NULL;
-		return;
+		return false;
 	}
 
 	clust_index = dict_table_get_first_index(node->table);
@@ -1210,16 +1267,21 @@ close_table:
 	ut_ad(!node->ref->info_bits);
 
 	if (node->update->info_bits & REC_INFO_MIN_REC_FLAG) {
+		if ((node->update->info_bits & ~REC_INFO_DELETED_FLAG)
+		    != REC_INFO_MIN_REC_FLAG) {
+			ut_ad("wrong info_bits in undo log record" == 0);
+			goto close_table;
+		}
 		/* This must be an undo log record for a subsequent
 		instant ALTER TABLE, extending the metadata record. */
 		ut_ad(clust_index->is_instant());
-		if (node->update->info_bits != REC_INFO_MIN_REC_FLAG) {
-			ut_ad(!"wrong info_bits in undo log record");
-			goto close_table;
-		}
-		node->update->info_bits = REC_INFO_METADATA;
-		const_cast<dtuple_t*>(node->ref)->info_bits
-			= REC_INFO_METADATA;
+		ut_ad(clust_index->table->instant
+		      || !(node->update->info_bits & REC_INFO_DELETED_FLAG));
+		node->ref = &trx_undo_metadata;
+		node->update->info_bits = (node->update->info_bits
+					   & REC_INFO_DELETED_FLAG)
+			? REC_INFO_METADATA_ALTER
+			: REC_INFO_METADATA_ADD;
 	}
 
 	if (!row_undo_search_clust_to_pcur(node)) {
@@ -1257,6 +1319,8 @@ close_table:
 				     (node->cmpl_info & UPD_NODE_NO_ORD_CHANGE)
 					? NULL : ptr);
 	}
+
+	return true;
 }
 
 /***********************************************************//**
@@ -1269,34 +1333,19 @@ row_undo_mod(
 	que_thr_t*	thr)	/*!< in: query thread */
 {
 	dberr_t	err;
-	ibool	dict_locked;
-
-	ut_ad(node != NULL);
-	ut_ad(thr != NULL);
-	ut_ad(node->state == UNDO_NODE_MODIFY);
-	ut_ad(node->trx->in_rollback);
-	ut_ad(!trx_undo_roll_ptr_is_insert(node->roll_ptr));
-
-	dict_locked = thr_get_trx(thr)->dict_operation_lock_mode == RW_X_LATCH;
-
 	ut_ad(thr_get_trx(thr) == node->trx);
+	const bool dict_locked = node->trx->dict_operation_lock_mode
+		== RW_X_LATCH;
 
-	row_undo_mod_parse_undo_rec(node, dict_locked);
-
-	if (node->table == NULL) {
-		/* It is already undone, or will be undone by another query
-		thread, or table was dropped */
-
-		node->state = UNDO_NODE_FETCH_NEXT;
-
-		return(DB_SUCCESS);
+	if (!row_undo_mod_parse_undo_rec(node, dict_locked)) {
+		return DB_SUCCESS;
 	}
 
 	node->index = dict_table_get_first_index(node->table);
 	ut_ad(dict_index_is_clust(node->index));
 
 	if (node->ref->info_bits) {
-		ut_ad(node->ref->info_bits == REC_INFO_METADATA);
+		ut_ad(node->ref->is_metadata());
 		goto rollback_clust;
 	}
 
@@ -1347,7 +1396,7 @@ rollback_clust:
 			/* Do not attempt to update statistics when
 			executing ROLLBACK in the InnoDB SQL
 			interpreter, because in that case we would
-			already be holding dict_sys->mutex, which
+			already be holding dict_sys.mutex, which
 			would be acquired when updating statistics. */
 			if (update_statistics && !dict_locked) {
 				dict_stats_update_if_needed(node->table,
